import numbers
import os
import tempfile

import h5py
import numpy as np

from dclab.rtdc_dataset import write


def cleanup(afile):
    # cleanup
    try:
        os.remove(afile)
    except OSError:
        pass


def test_bulk_scalar():
    data = {"area_um": np.linspace(100.7, 110.9, 100)}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_scalar_")
    write(rtdc_file, data)
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "area_um" in events.keys()
        assert np.all(events["area_um"][:] == data["area_um"])
    cleanup(rtdc_file)


def test_bulk_contour():
    num = 7
    contour = []
    for ii in range(5, num + 5):
        cii = np.arange(2 * ii).reshape(2, ii)
        contour.append(cii)
    data = {"area_um": np.linspace(100.7, 110.9, num),
            "contour": contour}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_contour_")
    write(rtdc_file, data)
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "contour" in events.keys()
        assert np.allclose(events["contour"]["6"], contour[6])
        assert events["contour"]["1"].shape == (2, 6)
    cleanup(rtdc_file)


def test_bulk_image():
    num = 7
    image = np.zeros((20, 90, 50))
    image += np.arange(90).reshape(1, 90, 1)
    data = {"area_um": np.linspace(100.7, 110.9, num),
            "image": image}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_image_")
    write(rtdc_file, data)
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "image" in events.keys()
        assert np.allclose(events["image"][6], image[6])
    cleanup(rtdc_file)


def test_bulk_mask():
    num = 7
    mask = []
    for ii in range(5, num + 5):
        mii = np.zeros(200, dtype=bool)
        mii[:ii] = True
        mask.append(mii.reshape(20, 10))
    data = {"area_um": np.linspace(100.7, 110.9, num),
            "mask": mask}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_mask_")
    write(rtdc_file, data)
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "mask" in events.keys()
        # Masks are stored as uint8
        assert np.allclose(events["mask"][6], mask[6]*255)
        assert events["mask"][1].shape == (20, 10)
    cleanup(rtdc_file)


def test_bulk_logs():
    log = ["This is a test log that contains two lines.",
           "This is the second line.",
           ]
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_logs_")
    write(rtdc_file, logs={"testlog": log})
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        outlog = rtdc_data["logs"]["testlog"]
        for ii in range(len(outlog)):
            assert outlog[ii] == log[ii]
    cleanup(rtdc_file)


def test_bulk_trace():
    num = 20
    trace = {"fl1_median": np.arange(num * 111).reshape(num, 111),
             "fl1_raw": 13 + np.arange(num * 111).reshape(num, 111),
             }
    data = {"area_um": np.linspace(100.7, 110.9, num),
            "trace": trace}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_trace_")
    write(rtdc_file, data)
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "trace" in events.keys()
        assert np.allclose(events["trace"]["fl1_raw"], trace["fl1_raw"])
    cleanup(rtdc_file)


def test_data_error():
    data = {"area_um": np.linspace(100.7, 110.9, 100)}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_")
    try:
        write(rtdc_file, data, mode="unknown")
    except ValueError:
        pass
    else:
        assert False, "ValueError should have been raised (wrong mode)"

    try:
        write(rtdc_file, ["peter"])
    except ValueError:
        pass
    else:
        assert False, "ValueError should have been raised (wrong data)"

    data2 = {"area_undefined": np.linspace(100.7, 110.9, 100)}
    try:
        write(rtdc_file, data2)
    except ValueError:
        pass
    else:
        assert False, "ValueError should have been raised (feature name)"

    data3 = {"trace": {"fl_unknown": np.arange(10)}}
    try:
        write(rtdc_file, data3)
    except ValueError:
        pass
    else:
        assert False, "ValueError should have been raised (trace name)"

    cleanup(rtdc_file)


def test_logs_append():
    log1 = ["This is a test log that contains two lines.",
            "This is the second line.",
            ]
    log2 = ["These are other logging events.",
            "They are appended to the log.",
            "And may have different lengths."
            ]
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_append_logs_")
    with h5py.File(rtdc_file, "w") as fobj:
        write(fobj, logs={"testlog": log1}, mode="append")
        write(fobj, logs={"testlog": log2}, mode="append")
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        outlog = rtdc_data["logs"]["testlog"]
        for ii in range(len(outlog)):
            assert outlog[ii] == (log1 + log2)[ii]
    cleanup(rtdc_file)


def test_meta():
    data = {"area_um": np.linspace(100.7, 110.9, 100)}
    meta = {"setup": {
        "channel width": 20,
        "chip region": "Channel",
    },
        "online_contour": {
        "no absdiff": "True",
        "image blur": 3.0,
    },
    }
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_meta_")
    write(rtdc_file, data, meta=meta)
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        abool = rtdc_data.attrs["online_contour:no absdiff"]
        assert abool
        assert isinstance(abool, (bool, np.bool_))
        anint = rtdc_data.attrs["online_contour:image blur"]
        assert isinstance(anint, numbers.Integral)
        assert rtdc_data.attrs["setup:channel width"] == 20
        assert rtdc_data.attrs["setup:chip region"] == "channel"
    cleanup(rtdc_file)


def test_meta_error():
    data = {"area_um": np.linspace(100.7, 110.9, 100)}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_meta_")

    meta1 = {"rediculous_section": {"a": 4}}
    try:
        write(rtdc_file, data, meta=meta1)
    except ValueError:
        pass
    else:
        assert False, "ValueError should have been raised (unkwn section)"

    meta2 = {"setup": {"rediculous_key": 4}}
    try:
        write(rtdc_file, data, meta=meta2)
    except ValueError:
        pass
    else:
        assert False, "ValueError should have been raised (unkwn key)"

    cleanup(rtdc_file)


def test_mode():
    data = {"area_um": np.linspace(100.7, 110.9, 100)}
    data2 = {"deform": np.linspace(.7, .8, 100)}

    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_replace_")
    write(rtdc_file, data=data, mode="reset")
    write(rtdc_file, data=data, mode="append").close()
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data1:
        events1 = rtdc_data1["events"]
        assert "area_um" in events1.keys()
        assert len(events1["area_um"]) == 2 * len(data["area_um"])

    write(rtdc_file, data=data, mode="replace")
    write(rtdc_file, data=data2, mode="replace")
    with h5py.File(rtdc_file, mode="r") as rtdc_data2:
        events2 = rtdc_data2["events"]
        assert "area_um" in events2.keys()
        assert "deform" in events2.keys()
        assert len(events2["area_um"]) == len(data["area_um"])
    cleanup(rtdc_file)


def test_mode_return():
    data = {"area_um": np.linspace(100.7, 110.9, 100)}

    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_return_")

    ret1 = write(rtdc_file, data=data, mode="append")
    assert isinstance(ret1, h5py.File)
    ret1.close()

    ret2 = write(rtdc_file, data=data, mode="replace")
    assert ret2 is None

    ret3 = write(rtdc_file, data=data, mode="reset")
    assert ret3 is None

    cleanup(rtdc_file)


def test_real_time():
    # Create huge array
    N = 116
    # Writing 10 images at a time is faster than writing one image at a time
    M = 4
    assert N // M == np.round(N / M)
    shx = 48
    shy = 32
    contours = [np.arange(20).reshape(10, 2)] * M
    images = np.zeros((M, shy, shx), dtype=np.uint8)
    masks = np.zeros((M, shy, shx), dtype=np.bool)
    traces = {"fl1_median": np.arange(M * 55).reshape(M, 55)}
    axis1 = np.linspace(0, 1, M)
    axis2 = np.arange(M)
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_realtime_")
    with h5py.File(rtdc_file, "w") as fobj:
        # simulate real time and write one image at a time
        for ii in range(N // M):
            # print(ii)
            num_img = np.copy(images) + ii

            data = {"area_um": axis1,
                    "area_cvx": axis2,
                    "image": num_img,
                    "contour": contours,
                    "mask": masks,
                    "trace": traces}

            write(fobj,
                  data=data,
                  mode="append",
                  )

    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert events["image"].shape == (N, shy, shx)
        assert events["area_um"].shape == (N,)
        assert events["contour"]["0"].shape == (10, 2)
        assert events["trace"]["fl1_median"].shape == (N, 55)
        assert np.dtype(events["area_um"]) == np.float
        assert np.dtype(events["area_cvx"]) == np.int

    cleanup(rtdc_file)


def test_real_time_single():
    # Create huge array
    N = 33
    shx = 30
    shy = 10
    image = np.zeros((shy, shx), dtype=np.uint8)
    mask = np.zeros((shy, shx), dtype=np.bool)
    contour = np.arange(22).reshape(11, 2)
    trace = {"fl1_median": np.arange(43)}

    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_realtime_single_")
    with h5py.File(rtdc_file, "w") as fobj:
        # simulate real time and write one image at a time
        for ii in range(N):
            data = {"area_um": ii * .1,
                    "area_cvx": ii * 5,
                    "image": image * ii,
                    "contour": contour,
                    "mask": mask,
                    "trace": trace}
            write(fobj,
                  data=data,
                  mode="append",
                  logs={"log1": "line {}".format(ii)}
                  )

    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert events["image"].shape == (N, shy, shx)
        assert events["area_um"].shape == (N,)
        assert events["contour"]["0"].shape == (11, 2)
        assert events["trace"]["fl1_median"].shape == (N, 43)
        assert np.dtype(events["area_um"]) == np.float
        assert np.dtype(events["area_cvx"]) == np.int
        logs = rtdc_data["logs"]
        assert len(logs["log1"]) == N

    cleanup(rtdc_file)


def test_replace_contour():
    num = 7
    contour = []
    contour2 = []
    for ii in range(5, num + 5):
        cii = np.arange(2 * ii).reshape(2, ii)
        contour.append(cii)
        contour2.append(cii * 2)

    data1 = {"area_um": np.linspace(100.7, 110.9, num),
             "contour": contour}
    data2 = {"contour": contour2}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_replace_contour_")
    write(rtdc_file, data1)
    write(rtdc_file, data2, mode="replace")
    # verify
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "contour" in events.keys()
        assert not np.allclose(events["contour"]["6"], contour[6])
        assert np.allclose(events["contour"]["6"], contour2[6])
    cleanup(rtdc_file)


def test_replace_logs():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_replace_logs_")
    write(rtdc_file, logs={"log1": ["hans", "und"]})
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        logs = rtdc_data["logs"]
        assert len(logs["log1"]) == 2

    write(rtdc_file, logs={"log1": ["peter"]}, mode="replace")
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        logs = rtdc_data["logs"]
        assert len(logs["log1"]) == 1
    cleanup(rtdc_file)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import numpy as np

from dclab.polygon_filter import PolygonFilter


def test_square_edges():
    poly = np.array([[0, 0],
                     [0, 1],
                     [1, 1],
                     [1, 0]])
    assert PolygonFilter.point_in_poly(p=(0, .5), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(1, .5), poly=poly)
    assert PolygonFilter.point_in_poly(p=(.5, 0), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(.5, 1), poly=poly)


def test_square_corners():
    poly = np.array([[0, 0],
                     [0, 1],
                     [1, 1],
                     [1, 0]])

    assert PolygonFilter.point_in_poly(p=(0, 0), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(1, 0), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(0, 1), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(1, 1), poly=poly)


def test_modsquare_edges_1():
    poly = np.array([[0, 0],
                     [0, 1],
                     [-1, 2],  # a dent to the top left
                     [1, 1],
                     [1, 0]])

    assert PolygonFilter.point_in_poly(p=(0, .5), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(1, .5), poly=poly)
    assert PolygonFilter.point_in_poly(p=(.5, 0), poly=poly)
    assert PolygonFilter.point_in_poly(p=(.5, 1), poly=poly)


def test_modsquare_edges_2():
    poly = np.array([[0, 0],
                     [0, 1],
                     [1, 1],
                     [2, -1],  # a dent to the bottom right
                     [1, 0],
                     ])

    assert PolygonFilter.point_in_poly(p=(0, .5), poly=poly)
    assert PolygonFilter.point_in_poly(p=(1, .5), poly=poly)
    assert PolygonFilter.point_in_poly(p=(.5, 0), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(.5, 1), poly=poly)


def test_modsquare_outside():
    poly = np.array([[0, 0],
                     [0, 1],
                     [-1, 2],  # a dent to the top left
                     [1, 1],
                     [1, 0]])

    assert PolygonFilter.point_in_poly(p=(.5, .5), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(-.5, .5), poly=poly)


def test_triangle_edges():
    poly = np.array([[0, 0],
                     [0, 1],
                     [1, 1],
                     ])

    assert PolygonFilter.point_in_poly(p=(0, .5), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(1, .5), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(.5, 0), poly=poly)
    assert not PolygonFilter.point_in_poly(p=(.5, 1), poly=poly)


if __name__ == "__main__":
    test_modsquare_outside()
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""RT-DC hierarchy format"""
from __future__ import division, print_function, unicode_literals

import numpy as np

from .. import definitions as dfn

from ..util import hashobj

from .config import Configuration
from .core import RTDCBase
from .filter import Filter


class HierarchyFilterError(BaseException):
    """Used for unexpected filtering operations"""
    pass


class ChildBase(object):
    def __init__(self, child):
        self.child = child

    def __len__(self):
        return len(self.child)


class ChildContour(ChildBase):
    def __getitem__(self, idx):
        pidx = map_indices_child2parent(child=self.child,
                                        child_indices=[idx])[0]
        hp = self.child.hparent
        return hp["contour"][pidx]


class ChildImage(ChildBase):
    def __getitem__(self, idx):
        pidx = map_indices_child2parent(child=self.child,
                                        child_indices=[idx])[0]
        hp = self.child.hparent
        return hp["image"][pidx]


class ChildMask(ChildBase):
    def __getitem__(self, idx):
        pidx = map_indices_child2parent(child=self.child,
                                        child_indices=[idx])[0]
        hp = self.child.hparent
        return hp["mask"][pidx]


class ChildTrace(ChildBase):
    def __init__(self, child, flname):
        super(ChildTrace, self).__init__(child)
        self.flname = flname

    def __getitem__(self, idx):
        pidx = map_indices_child2parent(child=self.child,
                                        child_indices=[idx])[0]
        hp = self.child.hparent
        return hp["trace"][self.flname][pidx]


class HierarchyFilter(Filter):
    def __init__(self, rtdc_ds, *args, **kwargs):
        """A filtering class for RTDC_Hierarchy

        This subclass handles manual filters for hierarchy children.
        The general problem with hierarchy children is that their data
        changes when the hierarchy parent changes. As hierarchy
        children may also have hierarchy children, dealing with
        manual filters (`Filter.manual`) is not trivial. Here,
        the manual filters are translated into event indices of the
        root parent (the highest member of the hierarchy, which is
        `RTDC_Hierarchy.hparent` if there is only one child).
        This enables to keep track of the manually excluded events
        even if

        - the parent changes its filters,
        - the parent is a hierarchy child as well, or
        - the excluded event is filtered out in the parent.
        """
        super(HierarchyFilter, self).__init__(rtdc_ds, *args, **kwargs)
        self.update_parent(rtdc_ds.hparent)
        self._man_root_ids = []

    @property
    def parent_changed(self):
        return hashobj(self._parent_rtdc_ds.filter.all) != self._parent_hash

    def apply_manual_indices(self, rtdc_ds, manual_indices):
        """Write to `self.manual`

        Write `manual_indices` to the boolean array `self.manual`
        and also store the indices as `self._man_root_ids`.

        Notes
        -----
        If `self.parent_changed` is `True`, i.e. the parent applied
        a filter and the child did not yet hear about this, then
        `HierarchyFilterError` is raised. This is important, because
        the size of the current filter would not match the size of
        the filtered events of the parent and thus index-mapping
        would not work.
        """
        if self.parent_changed:
            msg = "Cannot apply filter, because parent changed: " \
                  + "dataset {}. ".format(rtdc_ds) \
                  + "Run `RTDC_Hierarchy.apply_filter()` first!"
            raise HierarchyFilterError(msg)
        else:
            self._man_root_ids = list(manual_indices)
            cidx = map_indices_root2child(child=rtdc_ds,
                                          root_indices=manual_indices)
            if len(cidx):
                self.manual[cidx] = False

    def reset(self):
        super(HierarchyFilter, self).reset()
        self._man_root_ids = []

    def retrieve_manual_indices(self, rtdc_ds):
        """Read from self.manual

        Read from the boolean array `self.manual`, index all
        occurences of `False` and find the corresponding indices
        in the root hierarchy parent, return those and store them
        in `self._man_root_ids` as well.

        Notes
        -----
        This method also retrieves hidden indices, i.e. events
        that are not part of the current hierarchy child but
        which have been manually excluded before and are now
        hidden because a hierarchy parent filtered it out.

        If `self.parent_changed` is `True`, i.e. the parent applied
        a filter and the child did not yet hear about this, then
        nothing is computed and `self._man_root_ids` as-is.  This
        is important, because the size of the current filter would
        not match the size of the filtered events of the parent and
        thus index-mapping would not work.
        """
        if self.parent_changed:
            # ignore
            pass
        else:
            # indices from boolean array
            pbool = map_indices_child2root(
                child=rtdc_ds,
                child_indices=np.where(~self.manual)[0]).tolist()
            # retrieve all indices that are currently not visible
            # previous indices
            pold = self._man_root_ids
            # all indices previously selected either via
            # - self.manual or
            # - self.apply_manual_indices
            pall = sorted(list(set(pbool + pold)))
            # visible indices (only available child indices are returned)
            pvis_c = map_indices_root2child(child=rtdc_ds,
                                            root_indices=pall).tolist()
            # map visible child indices back to root indices
            pvis_p = map_indices_child2root(child=rtdc_ds,
                                            child_indices=pvis_c).tolist()
            # hidden indices
            phid = list(set(pall) - set(pvis_p))
            # Why not set `all_idx` to `pall`:
            # - pbool is considered to be correct
            # - pold contains hidden indices, but also might contain
            #   excess indices from before, i.e. if self.apply_manual_indices
            #   is called, self.manual is also updated. If however,
            #   self.manual is updated, self._man_root_ids are not updated.
            #   Thus, we trust pbool (self.manual) and only use pold
            #   (self._man_root_ids) to determine hidden indices.
            all_idx = list(set(pbool + phid))
            self._man_root_ids = sorted(all_idx)
        return self._man_root_ids

    def update_parent(self, parent_rtdc_ds):
        # hold reference to rtdc_ds parent
        # (not to its filter, because that is reinstantiated)
        self._parent_rtdc_ds = parent_rtdc_ds
        self._parent_hash = hashobj(self._parent_rtdc_ds.filter.all)


class RTDC_Hierarchy(RTDCBase):
    def __init__(self, hparent, apply_filter=True, *args, **kwargs):
        """Hierarchy dataset (filtered from RTDCBase)

        A few words on hierarchies:
        The idea is that a subclass of RTDCBase can use the filtered
        data of another subclass of RTDCBase and interpret these data
        as unfiltered events. This comes in handy e.g. when the
        percentage of different subpopulations need to be distinguished
        without the noise in the original data.

        Children in hierarchies always update their data according to
        the filtered event data from their parent when `apply_filter`
        is called. This makes it easier to save and load hierarchy
        children with e.g. Shape-Out and it makes the handling of
        hierarchies more intuitive (when the parent changes,
        the child changes as well).

        Parameters
        ----------
        hparent: instance of RTDCBase
            The hierarchy parent
        apply_filter: bool
            Whether to apply the filter durint instantiation;
            If set to `False`, `apply_filter` must be called
            manually.
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        hparent: RTDCBase
            Hierarchy parent of this instance
        """
        super(RTDC_Hierarchy, self).__init__(*args, **kwargs)

        self.path = hparent.path
        self.title = hparent.title + "_child"

        #: hierarchy parent
        self.hparent = hparent

        self.filter = HierarchyFilter(self)

        # Copy configuration
        cfg = hparent.config.copy()

        # Remove previously applied filters
        pops = []
        for key in cfg["filtering"]:
            if (key.endswith("min") or
                key.endswith("max") or
                    key == "polygon filters"):
                pops.append(key)

        [cfg["filtering"].pop(key) for key in pops]
        # Add parent information in dictionary
        cfg["filtering"]["hierarchy parent"] = hparent.identifier

        self.config = Configuration(cfg=cfg)

        if apply_filter:
            # Apply the filter
            # This will also populate all event attributes
            self.apply_filter()

    def __contains__(self, key):
        return self.hparent.__contains__(key)

    def __getitem__(self, key):
        # contour, image, and traces are added automatically
        # to `self._events` in `self.apply_filter`.
        if key not in self._events:
            item = self.hparent[key]
            if dfn.scalar_feature_exists(key):
                self._events[key] = item[self.hparent.filter.all]
        return self._events[key]

    def __len__(self):
        return np.sum(self.hparent.filter.all)

    def _check_parent_filter(self):
        """Reset filter if parent changed

        This will create a new HierarchyFilter for self if the
        parent RTDCBase changed. We do it like this, because it
        would be complicated to track all the changes in
        HierarchyFilter.
        """
        if self.filter.parent_changed:
            manual_pidx = self.filter.retrieve_manual_indices(self)
            self.filter = HierarchyFilter(self)
            self.filter.apply_manual_indices(self, manual_pidx)

    @property
    def features(self):
        return self.hparent.features

    @property
    def features_innate(self):
        return self.hparent.features_innate

    @property
    def features_loaded(self):
        return self.hparent.features_loaded

    @property
    def features_scalar(self):
        return self.hparent.features_scalar

    @property
    def hash(self):
        """Hashes of a hierarchy child changes if the parent changes"""
        # Do not apply filters here (speed)
        hph = self.hparent.hash
        hpfilt = hashobj(self.hparent.filter.all)
        dhash = hashobj(hph + hpfilt)
        return dhash

    def apply_filter(self, *args, **kwargs):
        """Overridden `apply_filter` to perform tasks for hierarchy child"""
        if self.filter is not None:
            # make sure self.filter knows about root manual indices
            # (stored in self.filter._man_root_ids)
            self.filter.retrieve_manual_indices(self)

        # Copy event data from hierarchy parent
        self.hparent.apply_filter(*args, **kwargs)
        # update event index
        event_count = np.sum(self.hparent.filter.all)
        self._events = {}
        self._events["index"] = np.arange(1, event_count + 1)
        # set non-scalar column data
        if "contour" in self.hparent:
            self._events["contour"] = ChildContour(self)
        if "image" in self.hparent:
            self._events["image"] = ChildImage(self)
        if "mask" in self.hparent:
            self._events["mask"] = ChildMask(self)
        if "trace" in self.hparent:
            trdict = {}
            for flname in dfn.FLUOR_TRACES:
                if flname in self.hparent["trace"]:
                    trdict[flname] = ChildTrace(self, flname)
            self._events["trace"] = trdict
        # update config
        self.config["experiment"]["event count"] = event_count

        # create a new filter if the parent changed
        self._check_parent_filter()
        super(RTDC_Hierarchy, self).apply_filter(*args, **kwargs)


def map_indices_child2parent(child, child_indices):
    """Map child RTDCBase event indices to parent RTDCBase

    Parameters
    ----------
    child: RTDC_Hierarchy
        hierarchy child with `child_indices`
    child_indices: 1d ndarray
        child indices to map

    Returns
    -------
    parent_indices: 1d ndarray
        hierarchy parent indices
    """
    parent = child.hparent
    # filters
    pf = parent.filter.all
    # indices corresponding to all child events
    idx = np.where(pf)[0]  # True means present in the child
    # indices corresponding to selected child events
    parent_indices = idx[child_indices]
    return parent_indices


def map_indices_child2root(child, child_indices):
    """Map RTDC_Hierarchy event indices to root RTDCBase

    Parameters
    ----------
    child: RTDC_Hierarchy
        RTDCBase hierarchy child
    child_indices: 1d ndarray
        child indices to map

    Returns
    -------
    root_indices: 1d ndarray
        hierarchy root indices
        (not necessarily the indices of `parent`)
    """
    while True:
        indices = map_indices_child2parent(child=child,
                                           child_indices=child_indices)
        if isinstance(child.hparent, RTDC_Hierarchy):
            child = child.hparent
            child_indices = indices
        else:
            break
    return indices


def map_indices_parent2child(child, parent_indices):
    """Map parent RTDCBase event indices to RTDC_Hierarchy

    Parameters
    ----------
    parent: RTDC_Hierarchy
        hierarchy child
    parent_indices: 1d ndarray
        hierarchy parent (`child.hparent`) indices to map

    Returns
    -------
    child_indices: 1d ndarray
        child indices
    """
    parent = child.hparent
    # filters
    pf = parent.filter.all
    # indices in child
    child_indices = []
    count = 0
    for ii in range(len(pf)):
        if pf[ii]:
            # only append indices if they exist in child
            if ii in parent_indices:
                # current child event count is the child index
                child_indices.append(count)
            # increment child event count
            count += 1

    return np.array(child_indices)


def map_indices_root2child(child, root_indices):
    """Map root RTDCBase event indices to child RTDCBase

    Parameters
    ----------
    parent: RTDCBase
        hierarchy parent of `child`.
    root_indices: 1d ndarray
        hierarchy root indices to map
        (not necessarily the indices of `parent`)

    Returns
    -------
    child_indices: 1d ndarray
        child indices
    """
    # construct hierarchy tree containing only RTDC_Hierarchy instances
    hierarchy = [child]
    while True:
        if isinstance(child.hparent, RTDC_Hierarchy):
            # the parent is a hierarchy tree
            hierarchy.append(child.hparent)
            child = child.hparent
        else:
            break

    indices = root_indices
    for hp in hierarchy[::-1]:  # reverse order
        # For each hierarchy parent, map the indices down the
        # hierarchy tree.
        indices = map_indices_parent2child(child=hp,
                                           parent_indices=indices)
    return indices



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import pathlib
import time
import sys

import numpy as np
import pytest

import dclab
from dclab.rtdc_dataset import ancillaries

from helper_methods import example_data_dict, retrieve_data, \
    example_data_sets, cleanup


def test_0basic():
    ds = dclab.new_dataset(retrieve_data(example_data_sets[1]))
    for cc in ['fl1_pos',
               'frame',
               'size_x',
               'size_y',
               'contour',
               'area_cvx',
               'circ',
               'image',
               'trace',
               'fl1_width',
               'nevents',
               'pos_x',
               'pos_y',
               'fl1_area',
               'fl1_max',
               ]:
        assert cc in ds

    # ancillaries
    for cc in ["deform",
               "area_um",
               "aspect",
               "frame",
               "index",
               "time",
               ]:
        assert cc in ds

    cleanup()


def test_0error():
    keys = ["circ"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    try:
        ds["unknown_column"]
    except KeyError:
        pass
    else:
        raise ValueError("Should have raised KeyError!")


def test_aspect():
    # Aspect ratio of the data
    ds = dclab.new_dataset(retrieve_data("rtdc_data_traces_video_bright.zip"))
    aspect = ds["aspect"]
    assert np.sum(aspect > 1) == 904
    assert np.sum(aspect < 1) == 48
    cleanup()


def test_area_ratio():
    ds = dclab.new_dataset(retrieve_data("rtdc_data_traces_video.zip"))
    comp_ratio = ds["area_ratio"]
    # The convex area is always >= the raw area
    assert np.all(comp_ratio >= 1)
    assert np.allclose(comp_ratio[0], 1.0196464)
    cleanup()


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.event_image.'
                            + 'InitialFrameMissingWarning')
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'ancillaries.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_brightness():
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("rtdc_data_traces_video_bright.zip"))
    # This is something low-level and should not be done in a script.
    # Remove the brightness columns from RTDCBase to force computation with
    # the image and contour columns.
    real_avg = ds._events.pop("bright_avg")
    real_sd = ds._events.pop("bright_sd")
    # This will cause a zero-padding warning:
    comp_avg = ds["bright_avg"]
    comp_sd = ds["bright_sd"]
    idcompare = ~np.isnan(comp_avg)
    # ignore first event (no image data)
    idcompare[0] = False
    assert np.allclose(real_avg[idcompare], comp_avg[idcompare])
    assert np.allclose(real_sd[idcompare], comp_sd[idcompare])
    cleanup()


def test_contour_basic():
    ds1 = dclab.new_dataset(retrieve_data("rtdc_data_hdf5_mask_contour.zip"))
    # export all data except for contour data
    features = ds1.features
    features.remove("contour")
    dspath = pathlib.Path(ds1.path)
    tempout = dspath.parent / (dspath.name + "without_contour.rtdc")
    ds1.export.hdf5(tempout, features=features)
    ds2 = dclab.new_dataset(tempout)

    for ii in range(len(ds1)):
        cin = ds1["contour"][ii]
        cout = ds2["contour"][ii]
        # simple presence test
        for ci in cin:
            assert ci in cout
        # order
        for ii in range(1, len(cin)):
            c2 = np.roll(cin, ii, axis=0)
            if np.all(c2 == cout):
                break
        else:
            assert False, "contours not matching, check orientation?"
    cleanup()


def test_deform():
    keys = ["circ"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    assert np.allclose(ds["deform"], 1 - ds["circ"])


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_emodulus_known_media():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus model": "elastic sphere",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    # known-media
    ds2 = dclab.new_dataset(ddict)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus model": "elastic sphere",
                                 "emodulus medium": "CellCarrier",
                                 "emodulus temperature": 23.0
                                 }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    assert np.allclose(ds["emodulus"], ds2["emodulus"], equal_nan=True,
                       rtol=0, atol=1e-15)
    # ancillary feature priority check
    for af in ancillaries.AncillaryFeature.get_instances("emodulus"):
        if af.method.__name__ == "compute_emodulus_legacy":
            assert af.is_available(ds)
        else:
            assert not af.is_available(ds)
        if af.method.__name__ == "compute_emodulus_known_media":
            assert af.is_available(ds2)
        else:
            assert not af.is_available(ds2)


@pytest.mark.skipif(sys.version_info < (3, 3),
                    reason="perf_counter requires python3.3 or higher")
@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_emodulus_legacy():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus model": "elastic sphere",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    t1 = time.perf_counter()
    assert "emodulus" in ds
    t2 = time.perf_counter()

    t3 = time.perf_counter()
    ds["emodulus"]
    t4 = time.perf_counter()
    assert t4 - t3 > t2 - t1


def test_emodulus_legacy_area():
    # computes "area_um" from "area_cvx"
    keys = ["area_cvx", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    # area can be computed from areapix
    ds.config["imaging"]["pixel size"] = .34
    assert "area_um" in ds
    assert "emodulus" not in ds, "not config for emodulus"
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus model": "elastic sphere",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    assert "emodulus" in ds


def test_emodulus_legacy_none():
    keys = ["area_msd", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    assert "emodulus" not in ds, "not config for emodulus"
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus model": "elastic sphere",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    assert "emodulus" not in ds, "column 'area_um' should be missing"


def test_emodulus_legacy_none2():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    assert "emodulus" not in ds, "not config for emodulus"
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    assert "emodulus" not in ds, "emodulus model should be missing"


@pytest.mark.skipif(sys.version_info < (3, 3),
                    reason="perf_counter requires python3.3 or higher")
@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_emodulus_legacy_viscosity_does_not_matter():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus model": "elastic sphere",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5  # irrelevant
                                }
    ds2 = dclab.new_dataset(ddict)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus medium": "CellCarrier",
                                 "emodulus model": "elastic sphere",
                                 "emodulus temperature": 23.0,
                                 "emodulus viscosity": 0.1  # irrelevant
                                 }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    assert np.allclose(ds["emodulus"], ds2["emodulus"], equal_nan=True,
                       rtol=0, atol=1e-15)


def test_emodulus_reservoir():
    """Reservoir measurements should not have emodulus"""
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus model": "elastic sphere",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    assert "emodulus" in ds
    ds2 = dclab.new_dataset(ddict)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus medium": "CellCarrier",
                                 "emodulus model": "elastic sphere",
                                 "emodulus temperature": 23.0,
                                 "emodulus viscosity": 0.5
                                 }
    ds2.config["setup"]["chip region"] = "reservoir"
    assert "emodulus" not in ds2


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_emodulus_temp_feat():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus model": "elastic sphere",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    ddict2 = example_data_dict(size=8472, keys=keys)
    ddict2["temp"] = 23.0 * np.ones(8472)
    # temp-feat
    ds2 = dclab.new_dataset(ddict2)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus medium": "CellCarrier",
                                 "emodulus model": "elastic sphere",
                                 }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    assert np.allclose(ds["emodulus"], ds2["emodulus"], equal_nan=True,
                       rtol=0, atol=6.5e-14)
    # ancillary feature priority check
    for af in ancillaries.AncillaryFeature.get_instances("emodulus"):
        if af.method.__name__ == "compute_emodulus_legacy":
            assert af.is_available(ds)
        else:
            assert not af.is_available(ds)
        if af.method.__name__ == "compute_emodulus_temp_feat":
            assert af.is_available(ds2)
        else:
            assert not af.is_available(ds2)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_emodulus_temp_feat_2():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus model": "elastic sphere",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    ddict2 = example_data_dict(size=8472, keys=keys)
    ddict2["temp"] = 23.0 * np.ones(8472)
    ddict2["temp"][0] = 23.5  # change first element
    # temp-feat
    ds2 = dclab.new_dataset(ddict2)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus medium": "CellCarrier",
                                 "emodulus model": "elastic sphere",
                                 }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    assert np.allclose(ds["emodulus"][1:], ds2["emodulus"][1:], equal_nan=True,
                       rtol=0, atol=6.5e-14)
    assert not np.allclose(ds["emodulus"][0], ds2["emodulus"][0])
    ds3 = dclab.new_dataset(ddict)
    ds3.config["setup"]["flow rate"] = 0.16
    ds3.config["setup"]["channel width"] = 30
    ds3.config["imaging"]["pixel size"] = .34
    ds3.config["calculation"] = {"emodulus medium": "CellCarrier",
                                 "emodulus model": "elastic sphere",
                                 "emodulus temperature": 23.5,
                                 "emodulus viscosity": 0.5
                                 }
    assert np.allclose(ds3["emodulus"][0], ds2["emodulus"][0], rtol=0,
                       atol=6e-14)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_emodulus_visc_only():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus model": "elastic sphere",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5  # irrelevant
                                }
    # visc-only
    ds2 = dclab.new_dataset(ddict)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    visc = dclab.features.emodulus.viscosity.get_viscosity(
        medium="CellCarrier",
        channel_width=30,
        flow_rate=0.16,
        temperature=23.0)
    ds2.config["calculation"] = {"emodulus model": "elastic sphere",
                                 "emodulus viscosity": visc
                                 }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    assert np.allclose(ds["emodulus"], ds2["emodulus"], equal_nan=True,
                       rtol=0, atol=1e-15)
    # ancillary feature priority check
    for af in ancillaries.AncillaryFeature.get_instances("emodulus"):
        if af.method.__name__ == "compute_emodulus_legacy":
            assert af.is_available(ds)
        else:
            assert not af.is_available(ds)
        if af.method.__name__ == "compute_emodulus_visc_only":
            assert af.is_available(ds2)
        else:
            assert not af.is_available(ds2)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_emodulus_visc_only_2():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    visc = dclab.features.emodulus.viscosity.get_viscosity(
        medium="CellCarrier",
        channel_width=30,
        flow_rate=0.16,
        temperature=23.0)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus medium": "other",
                                "emodulus model": "elastic sphere",
                                "emodulus temperature": 47.0,  # irrelevant
                                "emodulus viscosity": visc
                                }
    # visc-only
    ds2 = dclab.new_dataset(ddict)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus model": "elastic sphere",
                                 "emodulus viscosity": visc
                                 }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    assert np.allclose(ds["emodulus"], ds2["emodulus"], equal_nan=True,
                       rtol=0, atol=1e-15)


def test_fl_crosstalk_2chan():
    ds = dclab.new_dataset(retrieve_data("rtdc_data_traces_2flchan.zip"))
    # simple example
    analysis = {"calculation": {"crosstalk fl12": 0,
                                "crosstalk fl21": .1}}
    ds.config.update(analysis)
    # normalization is c11 = c22 = c33 = 1
    assert np.allclose(ds["fl2_max"], ds["fl2_max_ctc"])
    assert not np.allclose(ds["fl1_max"], ds["fl1_max_ctc"])
    # advanced example
    ct12 = .5
    ct21 = .3
    analysis2 = {"calculation": {"crosstalk fl12": ct12,
                                 "crosstalk fl21": ct21}}
    # AncillaryColumn uses hashes to check whether a particular calculation
    # was already performed. Thus, just updating the config will trigger
    # a new crosstalk correction once the data is requested.
    ds.config.update(analysis2)
    fl1_max = ds["fl1_max_ctc"] + ct21 * ds["fl2_max_ctc"]
    fl2_max = ds["fl2_max_ctc"] + ct12 * ds["fl1_max_ctc"]
    assert np.allclose(fl1_max, ds["fl1_max"])
    assert np.allclose(fl2_max, ds["fl2_max"])


def test_fl_crosstalk_3chanvs2chan():
    data = {"fl1_max": np.linspace(1, 1.1, 10),
            "fl2_max": np.linspace(0, 4.1, 10),
            "fl3_max": np.linspace(3, 2.5, 10),
            }
    ds = dclab.new_dataset(data)
    analysis = {"calculation": {"crosstalk fl12": .4,
                                "crosstalk fl21": .05,
                                }}
    ds.config.update(analysis)
    assert "fl2_max_ctc" in ds
    try:
        ds["fl2_max_ctc"]
    except ancillaries.af_fl_max_ctc.MissingCrosstalkMatrixElementsError:
        pass
    else:
        assert False, "Crosstalk correction from missing data should not work"
    # add missing matrix elements
    analysis = {"calculation": {"crosstalk fl13": .1,
                                "crosstalk fl23": .7,
                                "crosstalk fl31": .2,
                                "crosstalk fl32": .2,
                                }}
    ds.config.update(analysis)
    ds["fl1_max_ctc"]
    ds["fl2_max_ctc"]
    ds["fl3_max_ctc"]
    ds.config.update(analysis)


def test_fl_crosstalk_missing():
    data = {"fl1_max": np.linspace(1, 1.1, 10),
            "fl2_max": np.linspace(0, 4.1, 10),
            }
    ds = dclab.new_dataset(data)
    analysis = {"calculation": {"crosstalk fl12": .4,
                                }}
    ds.config.update(analysis)
    assert "fl2_max_ctc" not in ds


def test_fl_crosstalk_priority():
    data = {"fl1_max": np.linspace(1, 1.1, 10),
            "fl2_max": np.linspace(0, 4.1, 10),
            "fl3_max": np.linspace(3, 2.5, 10),
            }
    ds = dclab.new_dataset(data)
    analysis = {"calculation": {"crosstalk fl12": .4,
                                "crosstalk fl21": .05,
                                }}
    ds.config.update(analysis)
    av = ancillaries.AncillaryFeature.available_features(ds)
    avkeys = list(av.keys())
    assert "fl1_max_ctc" in avkeys
    assert "fl2_max_ctc" in avkeys
    assert "fl3_max_ctc" not in avkeys
    reqconf = [['calculation', ['crosstalk fl21', 'crosstalk fl12']]]
    assert av["fl1_max_ctc"].req_config == reqconf
    analysis = {"calculation": {"crosstalk fl13": .1,
                                "crosstalk fl23": .7,
                                "crosstalk fl31": .2,
                                "crosstalk fl32": .2,
                                }}
    ds.config.update(analysis)
    # If there are three fl features and the corresponding crosstalk
    # values, then we must always have triple crosstalk correction.
    av2 = ancillaries.AncillaryFeature.available_features(ds)
    av2keys = list(av2.keys())
    assert "fl1_max_ctc" in av2keys
    assert "fl2_max_ctc" in av2keys
    assert "fl3_max_ctc" in av2keys
    reqconf2 = [['calculation', ["crosstalk fl21",
                                 "crosstalk fl31",
                                 "crosstalk fl12",
                                 "crosstalk fl32",
                                 "crosstalk fl13",
                                 "crosstalk fl23"]]]
    assert av2["fl1_max_ctc"].req_config == reqconf2
    assert av2["fl2_max_ctc"].req_config == reqconf2
    assert av2["fl3_max_ctc"].req_config == reqconf2


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'ancillaries.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_inert_ratio_cvx():
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("rtdc_data_traces_video_bright.zip"))
    # This is something low-level and should not be done in a script.
    # Remove the brightness columns from RTDCBase to force computation with
    # the image and contour columns.
    real_ir = ds._events.pop("inert_ratio_cvx")
    # This will cause a zero-padding warning:
    comp_ir = ds["inert_ratio_cvx"]
    idcompare = ~np.isnan(comp_ir)
    # ignore first event (no image data)
    idcompare[0] = False
    assert np.allclose(real_ir[idcompare], comp_ir[idcompare])
    cleanup()


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'ancillaries.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_inert_ratio_prnc():
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("rtdc_data_traces_video_bright.zip"))
    # This will cause a zero-padding warning:
    prnc = ds["inert_ratio_prnc"]
    raw = ds["inert_ratio_raw"]
    idcompare = ~np.isnan(prnc)
    # ignore first event (no image data)
    idcompare[0] = False
    diff = (prnc - raw)[idcompare]
    # only compare the first valid event which seems to be quite close
    assert np.allclose(diff[0], 0, atol=1.2e-3, rtol=0)
    cleanup()


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'ancillaries.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_inert_ratio_raw():
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("rtdc_data_traces_video_bright.zip"))
    # This is something low-level and should not be done in a script.
    # Remove the brightness columns from RTDCBase to force computation with
    # the image and contour columns.
    real_ir = ds._events.pop("inert_ratio_raw")
    # This will cause a zero-padding warning:
    comp_ir = ds["inert_ratio_raw"]
    idcompare = ~np.isnan(comp_ir)
    # ignore first event (no image data)
    idcompare[0] = False
    assert np.allclose(real_ir[idcompare], comp_ir[idcompare])
    cleanup()


def test_ml_class_basic():
    data = {"ml_score_001": [.1, .3, .1, 0.01, .59],
            "ml_score_002": [.2, .1, .4, 0, .8],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    assert np.allclose(ds["ml_class"], [1, 0, 1, 0, 1])
    assert issubclass(ds["ml_class"].dtype.type, np.integer)


def test_ml_class_bad_feature():
    data = {"ml_score_0-1": [.1, .3, .1, 0.01, .59],
            }
    try:
        dclab.new_dataset(data)
    except ValueError:
        pass
    else:
        assert False, "This is not a valid feature name"


def test_ml_class_bad_score_max():
    data = {"ml_score_001": [.1, .3, 99, 0.01, .59],
            "ml_score_002": [.2, .1, .4, 0, .8],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    try:
        ds["ml_class"]
    except ValueError as e:
        assert "> 1" in e.args[0]
    else:
        assert False, "99 is not allowed"


def test_ml_class_bad_score_min():
    data = {"ml_score_001": [.1, .3, -.1, 0.01, .59],
            "ml_score_002": [.2, .1, .4, 0, .8],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    try:
        ds["ml_class"]
    except ValueError as e:
        assert "< 0" in e.args[0]
    else:
        assert False, "negative is not allowed"


def test_ml_class_bad_score_nan():
    data = {"ml_score_001": [.1, .3, np.nan, 0.01, .59],
            "ml_score_002": [.2, .1, .4, 0, .8],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    try:
        ds["ml_class"]
    except ValueError as e:
        assert "nan values" in e.args[0]
    else:
        assert False, "nan is not allowed"


def test_ml_class_single():
    data = {"ml_score_001": [.1, .3, .1, 0.01, .59],
            }
    ds = dclab.new_dataset(data)
    assert "ml_class" in ds
    assert np.allclose(ds["ml_class"], 0)


def test_time():
    ds = dclab.new_dataset(retrieve_data("rtdc_data_minimal.zip"))
    tt = ds["time"]
    assert tt[0] == 0
    assert np.allclose(tt[1], 0.0385)
    assert np.all(np.diff(tt) > 0)
    cleanup()


def test_volume():
    ds = dclab.new_dataset(retrieve_data("rtdc_data_minimal.zip"))
    vol = ds["volume"]
    # There are a lot of nans, because the contour is not given everywhere
    vol = vol[~np.isnan(vol)]
    assert np.allclose(vol[0], 574.60368907528346)
    assert np.allclose(vol[12], 1010.5669523203878)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Export RT-DC measurement data"""
from __future__ import division, print_function, unicode_literals

import codecs
import pathlib
import warnings

import h5py

from ..compat import PyImportError

try:
    import imageio
except PyImportError:
    IMAGEIO_AVAILABLE = False
else:
    IMAGEIO_AVAILABLE = True

try:
    import fcswrite
except PyImportError:
    FCSWRITE_AVAILABLE = False
else:
    FCSWRITE_AVAILABLE = True

import numpy as np

from .. import definitions as dfn
from .._version import version
from .write_hdf5 import write


class LimitingExportSizeWarning(UserWarning):
    pass


class Export(object):
    def __init__(self, rtdc_ds):
        """Export functionalities for RT-DC datasets"""
        self.rtdc_ds = rtdc_ds

    def avi(self, path, filtered=True, override=False):
        """Exports filtered event images to an avi file

        Parameters
        ----------
        path: str
            Path to a .avi file. The ending .avi is added automatically.
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.

        Notes
        -----
        Raises OSError if current dataset does not contain image data
        """
        if not IMAGEIO_AVAILABLE:
            raise PyImportError("Package `imageio` required for avi export!")
        path = pathlib.Path(path)
        ds = self.rtdc_ds
        # Make sure that path ends with .avi
        if path.suffix != ".avi":
            path = path.with_name(path.name + ".avi")
        # Check if file already exist
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(
                str(path).encode("ascii", "ignore")) +
                "Please use the `override=True` option.")
        # Start exporting
        if "image" in ds:
            # Open video for writing
            vout = imageio.get_writer(uri=path,
                                      format="FFMPEG",
                                      fps=25,
                                      codec="rawvideo",
                                      pixelformat="yuv420p",
                                      macro_block_size=None,
                                      ffmpeg_log_level="error")
            # write the filtered frames to avi file
            for evid in np.arange(len(ds)):
                # skip frames that were filtered out
                if filtered and not ds.filter.all[evid]:
                    continue
                image = ds["image"][evid]
                # Convert image to RGB
                image = image.reshape(image.shape[0], image.shape[1], 1)
                image = np.repeat(image, 3, axis=2)
                vout.append_data(image)
        else:
            msg = "No image data to export: dataset {} !".format(ds.title)
            raise OSError(msg)

    def fcs(self, path, features, meta_data={}, filtered=True, override=False):
        """Export the data of an RT-DC dataset to an .fcs file

        Parameters
        ----------
        mm: instance of dclab.RTDCBase
            The dataset that will be exported.
        path: str
            Path to an .fcs file. The ending .fcs is added automatically.
        features: list of str
            The features in the resulting .fcs file. These are strings
            that are defined by `dclab.definitions.scalar_feature_exists`,
            e.g. "area_cvx", "deform", "frame", "fl1_max", "aspect".
        meta_data: dict
            User-defined, optional key-value pairs that are stored
            in the primary TEXT segment of the FCS file; the version
            of dclab is stored there by default
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.

        Notes
        -----
        Due to incompatibility with the .fcs file format, all events with
        NaN-valued features are not exported.
        """
        if not FCSWRITE_AVAILABLE:
            raise PyImportError("Package `fcswrite` required for fcs export!")
        features = [c.lower() for c in features]
        ds = self.rtdc_ds

        path = pathlib.Path(path)
        # Make sure that path ends with .fcs
        if path.suffix != ".fcs":
            path = path.with_name(path.name + ".fcs")
        # Check if file already exist
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(
                str(path).encode("ascii", "ignore")) +
                "Please use the `override=True` option.")
        # Check that features are valid
        for c in features:
            if c not in ds.features_scalar:
                msg = "Invalid feature name: {}".format(c)
                raise ValueError(msg)

        # Collect the header
        chn_names = [dfn.get_feature_label(c, rtdc_ds=ds) for c in features]

        # Collect the data
        if filtered:
            data = [ds[c][ds.filter.all] for c in features]
        else:
            data = [ds[c] for c in features]

        data = np.array(data).transpose()
        meta_data["dclab version"] = version
        fcswrite.write_fcs(filename=str(path),
                           chn_names=chn_names,
                           data=data,
                           text_kw_pr=meta_data,
                           )

    def hdf5(self, path, features, filtered=True, override=False,
             compression="gzip"):
        """Export the data of the current instance to an HDF5 file

        Parameters
        ----------
        path: str
            Path to an .rtdc file. The ending .rtdc is added
            automatically.
        features: list of str
            The features in the resulting .rtdc file. These are strings
            that are defined by `dclab.definitions.feature_exists`, e.g.
            "area_cvx", "deform", "frame", "fl1_max", "image".
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.
        compression: str or None
            Compression method for "contour", "image", and "trace" data
            as well as logs; one of [None, "lzf", "gzip", "szip"].
        """
        path = pathlib.Path(path)
        # Make sure that path ends with .rtdc
        if not path.suffix == ".rtdc":
            path = path.parent / (path.name + ".rtdc")
        # Check if file already exists
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(path)
                          + "Please use the `override=True` option.")
        elif path.exists():
            path.unlink()

        meta = {}

        # only export configuration meta data (no user-defined config)
        for sec in dfn.CFG_METADATA:
            if sec in ["fmt_tdms"]:
                # ignored sections
                continue
            if sec in self.rtdc_ds.config:
                meta[sec] = self.rtdc_ds.config[sec].copy()

        if filtered:
            filtarr = self.rtdc_ds.filter.all
        else:
            filtarr = np.ones(len(self.rtdc_ds), dtype=bool)

        # check that all features have same length and use smallest
        # common length
        lengths = []
        for feat in features:
            if feat == "trace":
                for tr in list(self.rtdc_ds["trace"].keys()):
                    lengths.append(len(self.rtdc_ds["trace"][tr]))
            else:
                lengths.append(len(self.rtdc_ds[feat]))
        if not np.all(lengths == lengths[0]):
            lmin = np.min(lengths)
            lmax = np.max(lengths)
            nev_bef = np.sum(filtarr)
            filtarr[lmin:] = False
            nev_aft = np.sum(filtarr)
            if nev_bef != nev_aft:
                warnings.warn(
                    "Not all features have the same length! "
                    + "Limiting output event count to {} ".format(lmin)
                    + "(max {}) in '{}'.".format(lmax, path),
                    LimitingExportSizeWarning)

        # write meta data
        with write(path_or_h5file=path, meta=meta, mode="append") as h5obj:
            # write each feature individually
            for feat in features:
                hdf5_append(h5obj=h5obj,
                            rtdc_ds=self.rtdc_ds,
                            feat=feat,
                            compression=compression,
                            filtarr=filtarr)
            # update configuration
            hdf5_autocomplete_config(h5obj)

    def tsv(self, path, features, meta_data={}, filtered=True, override=False):
        """Export the data of the current instance to a .tsv file

        Parameters
        ----------
        path: str
            Path to a .tsv file. The ending .tsv is added automatically.
        features: list of str
            The features in the resulting .tsv file. These are strings
            that are defined by `dclab.definitions.scalar_feature_exists`,
            e.g. "area_cvx", "deform", "frame", "fl1_max", "aspect".
        meta_data: dict
            User-defined, optional key-value pairs that are stored
            at the beginning of the tsv file - one key-value pair is
            stored per line which starts with a hash. The version of
            dclab is stored there by default.
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.
        """
        features = [c.lower() for c in features]
        path = pathlib.Path(path)
        ds = self.rtdc_ds
        # Make sure that path ends with .tsv
        if path.suffix != ".tsv":
            path = path.with_name(path.name + ".tsv")
        # Check if file already exist
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(
                str(path).encode("ascii", "ignore")) +
                "Please use the `override=True` option.")
        # Check that features exist
        for c in features:
            if c not in ds.features_scalar:
                raise ValueError("Invalid feature name {}".format(c))
        meta_data["dclab version"] = version
        # Write BOM header
        with path.open("wb") as fd:
            fd.write(codecs.BOM_UTF8)
        # Open file
        with path.open("a", encoding="utf-8") as fd:
            # write meta data
            for key in sorted(meta_data.keys()):
                fd.write("# {}: {}\n".format(key, meta_data[key]))
            fd.write("#\n")
            # write header
            header1 = "\t".join([c for c in features])
            fd.write("# "+header1+"\n")
            labels = [dfn.get_feature_label(c, rtdc_ds=ds) for c in features]
            header2 = "\t".join(labels)
            fd.write("# "+header2+"\n")

        with path.open("ab") as fd:
            # write data
            if filtered:
                data = [ds[c][ds.filter.all] for c in features]
            else:
                data = [ds[c] for c in features]

            np.savetxt(fd,
                       np.array(data).transpose(),
                       fmt=str("%.10e"),
                       delimiter="\t")


def hdf5_append(h5obj, rtdc_ds, feat, compression, filtarr=None,
                time_offset=0):
    """Append feature data to an HDF5 file

    Parameters
    ----------
    h5obj: h5py.File
        Opened HDF5 file
    rtdc_ds: dclab.rtdc_dataset.RTDCBase
        Instance from which to obtain the data
    feat: str
        Valid feature name in `rtdc_ds`
    compression: str or None
        Compression method for "contour", "image", and "trace" data
        as well as logs; one of [None, "lzf", "gzip", "szip"].
    filtarr: None or 1d boolean np.ndarray
        Optional boolean array used for filtering. If set to
        `None`, all events are saved.
    time_offset: float
        This value will be added to the "time" and "frame" features
        (used for joining multiple measurements)

    Notes
    -----
    Please update the "experiment::event count" attribute manually.
    You may use :func:`hdf5_autocomplete_config` for that.
    """
    # optional array for filtering events
    if filtarr is None:
        filtarr = np.ones(len(rtdc_ds), dtype=bool)
    # total number of new events
    nev = np.sum(filtarr)
    # event-wise, because
    # - tdms-based datasets don't allow indexing with numpy
    # - there might be memory issues
    if feat == "contour":
        cont_list = []
        cmax = 0
        for ii in range(len(rtdc_ds)):
            if filtarr[ii]:
                dat = rtdc_ds["contour"][ii]
                cont_list.append(dat)
                cmax = max(cmax, dat.max())
        write(h5obj,
              data={"contour": cont_list},
              mode="append",
              compression=compression)
    elif feat in ["mask", "image"]:
        # store image stacks (reduced file size and save time)
        m = 64
        im0 = rtdc_ds[feat][0]
        imstack = np.zeros((m, im0.shape[0], im0.shape[1]),
                           dtype=im0.dtype)
        jj = 0
        for ii in range(len(rtdc_ds)):
            if filtarr[ii]:
                dat = rtdc_ds[feat][ii]
                imstack[jj] = dat
                if (jj + 1) % m == 0:
                    jj = 0
                    write(h5obj,
                          data={feat: imstack},
                          mode="append",
                          compression=compression)
                else:
                    jj += 1
        # write rest
        if jj:
            write(h5obj,
                  data={feat: imstack[:jj, :, :]},
                  mode="append",
                  compression=compression)
    elif feat == "trace":
        for tr in rtdc_ds["trace"].keys():
            tr0 = rtdc_ds["trace"][tr][0]
            trdat = np.zeros((nev, tr0.size), dtype=tr0.dtype)
            jj = 0
            for ii in range(len(rtdc_ds)):
                if filtarr[ii]:
                    trdat[jj] = rtdc_ds["trace"][tr][ii]
                    jj += 1
            write(h5obj,
                  data={"trace": {tr: trdat}},
                  mode="append",
                  compression=compression)
    elif feat == "index":
        # re-enumerate data index feature (filtered data)
        if "events/index" in h5obj:
            nev0 = len(h5obj["events/index"])
        else:
            nev0 = 0
        write(h5obj,
              data={"index": np.arange(nev0+1, nev0+nev+1)},
              mode="append",
              compression=compression)
    elif feat == "index_online":
        if "events/index_online" in h5obj:
            # index_online is usually larger than index
            ido0 = h5obj["events/index_online"][-1] + 1
        else:
            ido0 = 0
        write(h5obj,
              data={"index_online": rtdc_ds["index_online"][filtarr] + ido0},
              mode="append",
              compression=compression)
    elif feat == "time":
        write(h5obj,
              data={"time": rtdc_ds["time"][filtarr] + time_offset},
              mode="append",
              compression=compression)
    elif feat == "frame":
        if time_offset != 0:
            # Only get the frame rate when we actually need it.
            fr = rtdc_ds.config["imaging"]["frame rate"]
        else:
            fr = 0
        frame_offset = time_offset * fr
        write(h5obj,
              data={"frame": rtdc_ds["frame"][filtarr] + frame_offset},
              mode="append",
              compression=compression)
    else:
        write(h5obj,
              data={feat: rtdc_ds[feat][filtarr]},
              mode="append",
              compression=compression)


def hdf5_autocomplete_config(path_or_h5obj):
    """"Autocompletes the configuration of the RTDC-measurement

    The following configuration keys are updated:

    - experiment:event count
    - fluorescence:samples per event
    - imaging: roi size x (if image or mask is given)
    - imaging: roi size y (if image or mask is given)

    The following configuration keys are added if not present:

    - fluorescence:channel count

    Parameters
    ----------
    path: pathlib.Path or str or h5py.File
        Path to or opened RT-DC measurement

    """
    if not isinstance(path_or_h5obj, h5py.File):
        close = True
        h5obj = h5py.File(path_or_h5obj, "a")
    else:
        close = False
        h5obj = path_or_h5obj

    # set event count
    feats = sorted(h5obj["events"].keys())
    if feats:
        h5obj.attrs["experiment:event count"] = len(h5obj["events"][feats[0]])
    else:
        raise ValueError("No features in '{}'!".format(path_or_h5obj))

    # set samples per event
    if "trace" in feats:
        traces = list(h5obj["events"]["trace"].keys())
        trsize = h5obj["events"]["trace"][traces[0]].shape[1]
        h5obj.attrs["fluorescence:samples per event"] = trsize

    # set channel count
    chcount = sum(["fl1_max" in feats, "fl2_max" in feats, "fl3_max" in feats])
    if chcount:
        if "fluorescence:channel count" not in h5obj.attrs:
            h5obj.attrs["fluorescence:channel count"] = chcount

    # set roi size x/y
    if "image" in h5obj["events"]:
        shape = h5obj["events"]["image"][0].shape
    elif "mask" in h5obj["events"]:
        shape = h5obj["events"]["mask"][0].shape
    else:
        shape = None
    if shape is not None:
        # update shape
        h5obj.attrs["imaging:roi size x"] = shape[1]
        h5obj.attrs["imaging:roi size y"] = shape[0]

    if close:
        h5obj.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Test DCOR format"""
from __future__ import print_function, unicode_literals

import socket
import sys

import dclab
from dclab.rtdc_dataset.fmt_dcor import RTDC_DCOR
import numpy as np
import pytest

from helper_methods import retrieve_data, cleanup


if sys.version_info[0] >= 3:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.connect(("dcor.mpl.mpg.de", 443))
            DCOR_AVAILABLE = True
        except socket.gaierror:
            DCOR_AVAILABLE = False
else:
    # skip test on python2
    DCOR_AVAILABLE = False


class MockAPIHandler(dclab.rtdc_dataset.fmt_dcor.APIHandler):
    def get(self, query, feat=None, trace=None, event=None):
        """Mocks communication with the DCOR API"""
        h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
        with dclab.new_dataset(h5path) as ds:
            if query == "size":
                return len(ds)
            elif query == "metadata":
                return ds.config
            elif query == "feature_list":
                return ds.features
            elif query == "feature" and dclab.dfn.scalar_feature_exists(feat):
                return ds[feat]
            elif query == "trace_list":
                return sorted(ds["trace"].keys())
            elif query == "trace":
                return ds["trace"][trace][event]
            else:
                return ds[feat][event]


def test_dcor_base(monkeypatch):
    monkeypatch.setattr(dclab.rtdc_dataset.fmt_dcor,
                        "APIHandler",
                        MockAPIHandler)
    with dclab.new_dataset(retrieve_data("rtdc_data_hdf5_rtfdc.zip")) as ds:
        dso = dclab.new_dataset("https://example.com/api/3/action/dcserv?id=1")
        assert len(dso) == len(ds)
        assert dso.config["setup"]["channel width"] == \
            ds.config["setup"]["channel width"]
        assert np.all(dso["area_um"] == ds["area_um"])
        assert np.all(dso["area_um"] == ds["area_um"])  # test cache
        assert np.all(dso["image"][4] == ds["image"][4])
        assert len(dso["image"]) == len(ds)
        for key in dso._events:
            assert key in ds
        for m, n in zip(dso["mask"], ds["mask"]):
            assert np.all(m == n)
        # compute an ancillary feature
        assert np.all(dso["volume"] == ds["volume"])
        assert np.all(dso["volume"] == ds["volume"])  # test cache
        # trace
        assert sorted(dso["trace"].keys()) == sorted(ds["trace"].keys())
        assert len(dso["trace"]["fl1_raw"]) == len(ds["trace"]["fl1_raw"])
        assert np.all(dso["trace"]["fl1_raw"][1] == ds["trace"]["fl1_raw"][1])
        for t1, t2 in zip(dso["trace"]["fl1_raw"], ds["trace"]["fl1_raw"]):
            assert np.all(t1 == t2)
    cleanup()


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="DCOR not reachable!")
def test_dcor_cache_scalar():
    # calibration beads
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        # sanity checks
        assert len(ds) == 5000
        assert "area_um" in ds

        area_um = ds["area_um"]
        assert ds["area_um"] is area_um, "Check proper caching"
        # provoke cache deletion
        ds._events._scalar_cache.pop("area_um")
        assert ds["area_um"] is not area_um, "test removal from cache"


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="DCOR not reachable!")
def test_dcor_cache_trace():
    # calibration beads
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        # sanity checks
        assert len(ds) == 5000
        assert "trace" in ds

        trace0 = ds["trace"]["fl1_raw"][0]
        assert ds["trace"]["fl1_raw"][0] is trace0, "Check proper caching"
        assert ds["trace"]["fl1_raw"][1] is not trace0, "Check proper caching"


def test_dcor_hierarchy(monkeypatch):
    monkeypatch.setattr(dclab.rtdc_dataset.fmt_dcor,
                        "APIHandler",
                        MockAPIHandler)
    dso = dclab.new_dataset("https://example.com/api/3/action/dcserv?id=1")
    dsh = dclab.new_dataset(dso)
    assert np.all(dso["area_um"] == dsh["area_um"])
    cleanup()


def test_url():
    target = "https://example.com/api/3/action/dcserv?id=123456"
    assert RTDC_DCOR.get_full_url(
        url="123456",
        use_ssl=True,
        host="example.com") == target
    assert RTDC_DCOR.get_full_url(
        url="http://example.com/api/3/action/dcserv?id=123456",
        use_ssl=True,
        host="example.com") == target
    assert RTDC_DCOR.get_full_url(
        url="example.com/api/3/action/dcserv?id=123456",
        use_ssl=True,
        host="example.com") == target
    assert RTDC_DCOR.get_full_url(
        url="https://example.com/api/3/action/dcserv?id=123456",
        use_ssl=None,
        host="example.com") == target
    assert RTDC_DCOR.get_full_url(
        url="123456",
        use_ssl=None,
        host="example.com") == target
    target2 = "http://example.com/api/3/action/dcserv?id=123456"
    assert RTDC_DCOR.get_full_url(
        url="example.com/api/3/action/dcserv?id=123456",
        use_ssl=False,
        host="example.com") == target2
    assert RTDC_DCOR.get_full_url(
        url="https://example.com/api/3/action/dcserv?id=123456",
        use_ssl=False,
        host="example.com") == target2
    assert RTDC_DCOR.get_full_url(
        url="http://example.com/api/3/action/dcserv?id=123456",
        use_ssl=None,
        host="example.com") == target2


if __name__ == "__main__":
    # Run all tests
    from inspect import signature
    loc = locals()
    for key in list(loc.keys()):
        if (key.startswith("test_")
            and hasattr(loc[key], "__call__")
                and "monkeypatch" not in signature(loc[key]).parameters):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Class for efficiently handling contour data"""
from __future__ import division, print_function, unicode_literals

import sys
import warnings

import numpy as np

from ...features import inert_ratio

from .exc import ContourIndexingError


class ContourVerificationWarning(UserWarning):
    pass


class ContourColumn(object):
    def __init__(self, rtdc_dataset):
        """A wrapper for ContourData that takes into account event offsets

        Event offsets appear when the first event that is recorded in the
        tdms files does not have a corresponding contour in the contour
        text file.
        """
        fname = self.find_contour_file(rtdc_dataset)
        if fname is None:
            self.identifier = None
        else:
            if sys.version_info[0] == 2:
                self.identifier = str(fname).decode("utf-8")
            else:
                self.identifier = str(fname)
        if fname is not None:
            self._contour_data = ContourData(fname)
            self._initialized = False
        else:
            self._contour_data = []
            # prevent `determine_offset` to be called
            self._initialized = True
        self.frame = rtdc_dataset["frame"]
        # if they are set, these features are used for verifying the contour
        self.pxfeat = {}

        if "area_msd" in rtdc_dataset:
            self.pxfeat["area_msd"] = rtdc_dataset["area_msd"]
        if "pixel size" in rtdc_dataset.config["imaging"]:
            px_size = rtdc_dataset.config["imaging"]["pixel size"]
            for key in ["pos_x", "pos_y", "size_x", "size_y"]:
                if key not in rtdc_dataset.features_innate:
                    # abort
                    self.pxfeat.clear()
                    break
                self.pxfeat[key] = rtdc_dataset[key] / px_size
        if "image" in rtdc_dataset:
            self.shape = rtdc_dataset["image"].shape
        else:
            self.shape = None
        self.event_offset = 0

    def __getitem__(self, idx):
        if not self._initialized:
            self.determine_offset()
        idnew = idx-self.event_offset
        cdata = None
        if idnew < 0:
            # No contour data
            cdata = np.zeros((2, 2), dtype=int)
        else:
            # Assign contour based on stored frame index
            frame_ist = self.frame[idx]
            # Do not only check the exact frame, but +/- 2 events around it
            for idn in [idnew, idnew-1, idnew+1, idnew-2, idnew+2]:
                # check frame
                try:
                    frame_soll = self._contour_data.get_frame(idn)
                except IndexError:
                    # reached end of file
                    continue
                if np.allclose(frame_soll, frame_ist, rtol=0):
                    cdata = self._contour_data[idn]
                    break
        if cdata is None and self.shape and self.pxfeat:  #
            # The frame is wrong, but the contour might be correct.
            # We check that by verifying several features.
            cdata2 = self._contour_data[idnew]
            cont = np.zeros((self.shape[1], self.shape[0]))
            cont[cdata2[:, 0], cdata2[:, 1]] = True
            mm = inert_ratio.cont_moments_cv(cdata2)

            if (np.allclose(self.pxfeat["size_x"][idx],
                            np.ptp(cdata2[:, 0]) + 1,
                            rtol=0, atol=1e-5)
                and np.allclose(self.pxfeat["size_y"][idx],
                                np.ptp(cdata2[:, 1]) + 1,
                                rtol=0, atol=1e-5)
                and np.allclose(mm["m00"],
                                self.pxfeat["area_msd"][idx],
                                rtol=0, atol=1e-5)
                # atol=6 for positions, because the original positions
                # are computed from the convex contour, which would be
                # computed using cv2.convexHull(cdata2).
                and np.allclose(self.pxfeat["pos_x"][idx],
                                mm["m10"]/mm["m00"],
                                rtol=0, atol=6)
                and np.allclose(self.pxfeat["pos_y"][idx],
                                mm["m01"]/mm["m00"],
                                rtol=0, atol=6)):
                cdata = cdata2

        if cdata is None:
            # No idea what went wrong, but we make the beste guess and
            # issue a warning.
            cdata = self._contour_data[idnew]
            frame_c = self._contour_data.get_frame(idnew)
            warnings.warn(
                "Couldn't verify contour {} in {}".format(idx, self.identifier)
                + " (frame index {})!".format(frame_c),
                ContourVerificationWarning
            )
        return cdata

    def __len__(self):
        length = len(self._contour_data)
        if length:
            if not self._initialized:
                self.determine_offset()
            length += self.event_offset
        return length

    def determine_offset(self):
        """Determines the offset of the contours w.r.t. other data columns

        Notes
        -----
        - the "frame" column of `rtdc_dataset` is compared to
          the first contour in the contour text file to determine an
          offset by one event
        - modifies the property `event_offset` and sets `_initialized`
          to `True`
        """
        # In case of regular RTDC, the first contour is
        # missing. In case of fRTDC, it is there, so we
        # might have an offset. We find out if the first
        # contour frame is missing by comparing it to
        # the "frame" column of the rtdc dataset.
        fref = self._contour_data.get_frame(0)
        f0 = self.frame[0]
        f1 = self.frame[1]
        # Use allclose to avoid float/integer comparison problems
        if np.allclose(fref, f0, rtol=0):
            self.event_offset = 0
        elif np.allclose(fref, f1, rtol=0):
            self.event_offset = 1
        else:
            msg = "Contour data has unknown offset (frame {})!".format(fref)
            raise ContourIndexingError(msg)
        self._initialized = True

    @staticmethod
    def find_contour_file(rtdc_dataset):
        """Tries to find a contour file that belongs to an RTDC dataset

        Returns None if no contour file is found.
        """
        cont_id = rtdc_dataset.path.stem
        cands = [c.name for c in rtdc_dataset._fdir.glob("*_contours.txt")]
        cands = sorted(cands)
        # Search for perfect matches, e.g.
        # - M1_0.240000ul_s.tdms
        # - M1_0.240000ul_s_contours.txt
        for c1 in cands:
            if c1.startswith(cont_id):
                cfile = rtdc_dataset._fdir / c1
                break
        else:
            # Search for M* matches with most overlap, e.g.
            # - M1_0.240000ul_s.tdms
            # - M1_contours.txt
            for c2 in cands:
                if (c2.split("_")[0] == rtdc_dataset._mid):
                    # Do not confuse with M10_contours.txt
                    cfile = rtdc_dataset._fdir / c2
                    break
            else:
                cfile = None
        return cfile


class ContourData(object):
    def __init__(self, fname):
        """Access an MX_contour.txt as a dictionary

        Initialize this class with a *_contour.txt file.
        The individual contours can be accessed like a
        list (enumerated from 0 on).
        """
        self._initialized = False
        self.filename = fname

    def __getitem__(self, idx):
        cont = self.data[idx]
        cont = cont.strip()
        cont = cont.replace(")", "")
        cont = cont.replace("(", "")
        cont = cont.replace("(", "")
        cont = cont.replace("\n", ",")
        cont = cont.replace("   ", " ")
        cont = cont.replace("  ", " ")
        if len(cont) > 1:
            _frame, cont = cont.split(" ", 1)
            cont = cont.strip(" ,")
            data = np.fromstring(cont, sep=",", dtype=np.uint16).reshape(-1, 2)
            return data

    def __len__(self):
        return len(self.data)

    def _index_file(self):
        """Open and index the contour file

        This function populates the internal list of contours
        as strings which will be available as `self.data`.
        """
        with self.filename.open() as fd:
            data = fd.read()

        ident = "Contour in frame"
        self._data = data.split(ident)[1:]
        self._initialized = True

    @property
    def data(self):
        """Access self.data
        If `self._index_file` has not been computed before, this
        property will cause it to do so.
        """
        if not self._initialized:
            self._index_file()
        return self._data

    def get_frame(self, idx):
        """Return the frame number of a contour"""
        cont = self.data[idx]
        # previously was split using " ", but "(" is more general
        frame = int(cont.strip().split("(", 1)[0])
        return frame



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

from pkg_resources import resource_filename

import numpy as np
import pytest

from dclab.features import emodulus


def test_bad_lut_data():
    try:
        emodulus.load_lut("bad_string_asdkubhasd")
    except ValueError:
        pass
    else:
        assert False, "Invalid `lut_data` results in ValueError"


@pytest.mark.filterwarnings(
    'ignore::dclab.features.emodulus.KnowWhatYouAreDoingWarning')
def test_extrapolate():
    """Test whether spline interpolation gives reasonable results"""
    lut, _ = emodulus.load_lut("FEM-2Daxis")

    area_norm = lut[:, 0].max()
    emodulus.normalize(lut[:, 0], area_norm)

    deform_norm = lut[:, 1].max()
    emodulus.normalize(lut[:, 1], deform_norm)

    np.random.seed(47)
    more_than_5perc = []
    valid_ones = 0

    for _ in range(100):
        # pick a few values from the LUT
        ids = np.random.randint(0, lut.shape[0], 10)
        area_um = lut[ids, 0]
        deform = lut[ids, 1]
        # set the emodulus to zero
        emod = np.nan * np.zeros(deform.size)
        # "extrapolate" within the grid using the spline
        emodulus.extrapolate_emodulus(
            lut=lut,
            datax=area_um,
            deform=deform,
            emod=emod,
            deform_norm=deform_norm,
            inplace=True)
        valid = ~np.isnan(emod)
        valid_ones += np.sum(valid)
        res = np.abs(lut[ids, 2] - emod)[valid]/lut[ids, 2][valid]
        if np.sum(res > .05):
            more_than_5perc.append([ids, res])

    assert len(more_than_5perc) == 0
    assert valid_ones == 149


def test_load_lut_from_array():
    ref_lut, ref_meta = emodulus.load_lut("FEM-2Daxis")
    lut2, meta2 = emodulus.load_lut((ref_lut, ref_meta))
    assert np.all(ref_lut == lut2)
    assert ref_meta == meta2
    assert ref_lut is not lut2, "data should be copied"
    assert ref_meta is not meta2, "meta data should be copied"


def test_load_lut_from_path():
    ref_lut, ref_meta = emodulus.load_lut("FEM-2Daxis")
    path = resource_filename("dclab.features.emodulus",
                             emodulus.INTERNAL_LUTS["FEM-2Daxis"])
    lut2, meta2 = emodulus.load_lut(path)
    assert np.all(ref_lut == lut2)
    assert ref_meta == meta2


def test_load_lut_from_badobject():
    try:
        emodulus.load_lut({"test": "nonesense"})
    except ValueError:
        pass
    else:
        assert False, "dict should not be supported"


def test_load_lut_from_badpath():
    try:
        emodulus.load_lut("peter/pan.txt")
    except ValueError:
        pass
    else:
        assert False, "dict should not be supported"


def test_pixelation_correction_volume():
    ddelt = emodulus.get_pixelation_delta(feat_corr="deform",
                                          feat_absc="volume",
                                          data_absc=100,
                                          px_um=0.34)
    assert np.allclose(ddelt, 0.011464479831134636)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_simple_emod():
    x = np.linspace(0, 250, 100)
    y = np.linspace(0, 0.1, 100)
    x, y = np.meshgrid(x, y)

    emod = emodulus.get_emodulus(area_um=x,
                                 deform=y,
                                 medium="CellCarrier",
                                 channel_width=30,
                                 flow_rate=0.16,
                                 px_um=0,  # without pixelation correction
                                 temperature=23)

    assert np.allclose(emod[10, 50], 1.1875799054283109)
    assert np.allclose(emod[50, 50], 0.5527978323425833)
    assert np.allclose(emod[80, 50], 0.45677680242729324)

    assert np.allclose(emod[10, 80], 1.5744560306483262)
    assert np.allclose(emod[50, 80], 0.73534561544655519)
    assert np.allclose(emod[80, 80], 0.60737083178222251)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""RT-DC dictionary format"""
from __future__ import division, print_function, unicode_literals

import time

import numpy as np

from .. import definitions as dfn
from ..util import hashobj

from .config import Configuration
from .core import RTDCBase


class RTDC_Dict(RTDCBase):
    def __init__(self, ddict, *args, **kwargs):
        """Dictionary-based RT-DC dataset

        Parameters
        ----------
        ddict: dict
            Dictionary with features as keys (valid features like
            "area_cvx", "deform", "image" are defined by
            `dclab.definitions.feature_exists`) with which the class
            will be instantiated. The configuration is set to the
            default configuration of dclab.

            .. versionchanged:: 0.27.0
                Scalar features are automatically converted to arrays.
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`
        """
        assert ddict

        super(RTDC_Dict, self).__init__(*args, **kwargs)

        t = time.localtime()

        # Get an identifying string
        keys = list(ddict.keys())
        keys.sort()
        ids = hashobj(ddict[keys[0]])
        self._ids = ids
        self.path = "none"
        self.title = "{}_{:02d}_{:02d}/{}.dict".format(t[0], t[1], t[2], ids)

        # Populate events
        self._events = {}
        for key in ddict:
            if dfn.feature_exists(key):
                if dfn.scalar_feature_exists(key):
                    data = np.array(ddict[key])
                else:
                    data = ddict[key]
            else:
                raise ValueError("Invalid feature name '{}'".format(key))
            self._events[key] = data

        event_count = len(ddict[list(ddict.keys())[0]])

        self.config = Configuration()
        self.config["experiment"]["event count"] = event_count
        # Set up filtering
        self._init_filters()

    @property
    def hash(self):
        return self._ids



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import numpy as np

from dclab.features.fl_crosstalk import correct_crosstalk


def test_simple_crosstalk():
    fl1 = np.array([1.1, 3.1, 6.3])
    fl2 = np.array([1.6, 30.1, 16.3])
    fl3 = np.array([10.3, 7.1, 8.9])

    ct21 = .1
    ct31 = .5
    ct12 = .03
    ct32 = .25
    ct13 = .01
    ct23 = .2

    ct11 = 1
    ct22 = 1
    ct33 = 1

    # compute cross-talked data
    fl1_bleed = ct11 * fl1 + ct21 * fl2 + ct31 * fl3
    fl2_bleed = ct12 * fl1 + ct22 * fl2 + ct32 * fl3
    fl3_bleed = ct13 * fl1 + ct23 * fl2 + ct33 * fl3

    # obtain crosstalk-corrected data
    fl1_ctc = correct_crosstalk(fl1=fl1_bleed,
                                fl2=fl2_bleed,
                                fl3=fl3_bleed,
                                fl_channel=1,
                                ct21=ct21,
                                ct31=ct31,
                                ct12=ct12,
                                ct32=ct32,
                                ct13=ct13,
                                ct23=ct23)

    fl2_ctc = correct_crosstalk(fl1=fl1_bleed,
                                fl2=fl2_bleed,
                                fl3=fl3_bleed,
                                fl_channel=2,
                                ct21=ct21,
                                ct31=ct31,
                                ct12=ct12,
                                ct32=ct32,
                                ct13=ct13,
                                ct23=ct23)

    fl3_ctc = correct_crosstalk(fl1=fl1_bleed,
                                fl2=fl2_bleed,
                                fl3=fl3_bleed,
                                fl_channel=3,
                                ct21=ct21,
                                ct31=ct31,
                                ct12=ct12,
                                ct32=ct32,
                                ct13=ct13,
                                ct23=ct23)

    assert np.allclose(fl1, fl1_ctc)
    assert np.allclose(fl2, fl2_ctc)
    assert np.allclose(fl3, fl3_ctc)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Crosstalk-correction for fluorescence data"""
from __future__ import division, print_function, unicode_literals

import numpy as np


def get_compensation_matrix(ct21, ct31, ct12, ct32, ct13, ct23):
    """Compute crosstalk inversion matrix

    The spillover matrix is

    | | c11 c12 c13 |
    | | c21 c22 c23 |
    | | c31 c32 c33 |

    The diagonal elements are set to 1, i.e.

    ct11 = c22 = c33 = 1

    Parameters
    ----------
    cij: float
        Spill from channel i to channel j

    Returns
    -------
    inv: np.ndarray
        Compensation matrix (inverted spillover matrix)
    """
    ct11 = 1
    ct22 = 1
    ct33 = 1

    if ct21 < 0:
        raise ValueError("ct21 matrix element must not be negative!")

    if ct31 < 0:
        raise ValueError("ct31 matrix element must not be negative!")

    if ct12 < 0:
        raise ValueError("ct12 matrix element must not be negative!")

    if ct32 < 0:
        raise ValueError("ct32 matrix element must not be negative!")

    if ct13 < 0:
        raise ValueError("ct13 matrix element must not be negative!")

    if ct23 < 0:
        raise ValueError("ct23 matrix element must not be negative!")

    crosstalk = np.array([[ct11, ct12, ct13],
                          [ct21, ct22, ct23],
                          [ct31, ct32, ct33],
                          ])
    return np.linalg.inv(crosstalk)


def correct_crosstalk(fl1, fl2, fl3, fl_channel,
                      ct21=0, ct31=0, ct12=0, ct32=0, ct13=0, ct23=0):
    """Perform crosstalk correction

    Parameters
    ----------
    fli: int, float, or np.ndarray
        Measured fluorescence signals
    fl_channel: int (1, 2, or 3)
        The channel number for which the crosstalk-corrected signal
        should be computed
    cij: float
        Spill (crosstalk or bleed-through) from channel i to channel j
        This spill is computed from the fluorescence signal of e.g.
        single-stained positive control cells; It is defined by the
        ratio of the fluorescence signals of the two channels, i.e
        cij = flj / fli.

    See Also
    --------
    get_compensation_matrix: compute the inverse crosstalk matrix

    Notes
    -----
    If there are only two channels (e.g. fl1 and fl2), then the
    crosstalk to and from the other channel (ct31, ct32, ct13, ct23)
    should be set to zero.
    """
    fl_channel = int(fl_channel)
    if fl_channel not in [1, 2, 3]:
        raise ValueError("`fl_channel` must be 1, 2, or 3!")

    minv = get_compensation_matrix(ct21=ct21, ct31=ct31, ct12=ct12,
                                   ct32=ct32, ct13=ct13, ct23=ct23)

    col = minv[:, fl_channel - 1].flatten()
    flout = col[0] * fl1 + col[1] * fl2 + col[2] * fl3
    return flout



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""RT-DC file format writer"""
from __future__ import unicode_literals

import copy

import h5py
import numpy as np

from ..compat import hdf5_str
from .. import definitions as dfn
from .._version import version


def store_contour(h5group, data, compression):
    if not isinstance(data, (list, tuple)):
        # single event
        data = [data]
    grp = h5group.require_group("contour")
    curid = len(grp.keys())
    for ii, cc in enumerate(data):
        grp.create_dataset("{}".format(curid + ii),
                           data=cc,
                           fletcher32=True,
                           compression=compression)


def store_image(h5group, data, compression):
    if len(data.shape) == 2:
        # single event
        data = data.reshape(1, data.shape[0], data.shape[1])
    if "image" not in h5group:
        maxshape = (None, data.shape[1], data.shape[2])
        chunks = (100, data.shape[1], data.shape[2])
        dset = h5group.create_dataset("image",
                                      data=data,
                                      dtype=np.uint8,
                                      maxshape=maxshape,
                                      chunks=chunks,
                                      fletcher32=True,
                                      compression=compression)
        # Create and Set image attributes:
        # HDFView recognizes this as a series of images.
        # Use np.string_ as per
        # http://docs.h5py.org/en/stable/strings.html#compatibility
        dset.attrs.create('CLASS', np.string_('IMAGE'))
        dset.attrs.create('IMAGE_VERSION', np.string_('1.2'))
        dset.attrs.create('IMAGE_SUBCLASS', np.string_('IMAGE_GRAYSCALE'))
    else:
        dset = h5group["image"]
        oldsize = dset.shape[0]
        dset.resize(oldsize + data.shape[0], axis=0)
        dset[oldsize:] = data


def store_mask(h5group, data, compression):
    # store binary mask data as uint8 to allow visualization in HDFView
    data = np.asarray(data, dtype=np.uint8)
    if data.max() != 255 and data.max() != 0 and data.min() == 0:
        data = data / data.max() * 255
    if len(data.shape) == 2:
        # single event
        data = data.reshape(1, data.shape[0], data.shape[1])
    if "mask" not in h5group:
        maxshape = (None, data.shape[1], data.shape[2])
        chunks = (100, data.shape[1], data.shape[2])
        dset = h5group.create_dataset("mask",
                                      data=data,
                                      dtype=np.uint8,
                                      maxshape=maxshape,
                                      chunks=chunks,
                                      fletcher32=True,
                                      compression=compression)
        # Create and Set image attributes
        # HDFView recognizes this as a series of images
        dset.attrs.create('CLASS', np.string_('IMAGE'))
        dset.attrs.create('IMAGE_VERSION', np.string_('1.2'))
        dset.attrs.create('IMAGE_SUBCLASS', np.string_('IMAGE_GRAYSCALE'))
    else:
        dset = h5group["mask"]
        oldsize = dset.shape[0]
        dset.resize(oldsize + data.shape[0], axis=0)
        dset[oldsize:] = data


def store_scalar(h5group, name, data, compression):
    if np.isscalar(data):
        # single event
        data = np.atleast_1d(data)
    if name not in h5group:
        h5group.create_dataset(name,
                               data=data,
                               maxshape=(None,),
                               chunks=(100,),
                               fletcher32=True,
                               compression=compression
                               )
    else:
        dset = h5group[name]
        oldsize = dset.shape[0]
        dset.resize(oldsize + data.shape[0], axis=0)
        dset[oldsize:] = data


def store_trace(h5group, data, compression):
    firstkey = sorted(list(data.keys()))[0]
    if len(data[firstkey].shape) == 1:
        # single event
        for dd in data:
            data[dd] = data[dd].reshape(1, -1)
    # create trace group
    grp = h5group.require_group("trace")

    for flt in data:
        # create traces datasets
        if flt not in grp:
            maxshape = (None, data[flt].shape[1])
            chunks = (100, data[flt].shape[1])
            grp.create_dataset(flt,
                               data=data[flt],
                               maxshape=maxshape,
                               chunks=chunks,
                               fletcher32=True,
                               compression=compression)
        else:
            dset = grp[flt]
            oldsize = dset.shape[0]
            dset.resize(oldsize + data[flt].shape[0], axis=0)
            dset[oldsize:] = data[flt]


def write(path_or_h5file, data={}, meta={}, logs={}, mode="reset",
          compression=None):
    """Write data to an RT-DC file

    Parameters
    ----------
    path: path or h5py.File
        The path or the hdf5 file object to write to.
    data: dict-like
        The data to store. Each key of `data` must be a valid
        feature name (see :func:`dclab.dfn.feature_exists`). The
        data type must be given according to the feature type:

        - scalar feature: 1d ndarray of size `N`, any dtype,
          with the number of events `N`.
        - contour: list of `N` 2d ndarrays of shape `(2,C)`, any dtype,
          with each ndarray containing the x- and y- coordinates
          of `C` contour points in pixels.
        - image: 3d ndarray of shape `(N,A,B)`, is converted to uint8,
          with the image dimensions `(x,y) = (A,B)`
        - mask: 3d ndarray of shape `(N,A,B)`, is converted to bool,
          with the image dimensions `(x,y) = (A,B)`
        - trace: 2d ndarray of shape `(N,T)`, any dtype
          with a globally constant trace length `T`.
    meta: dict of dicts
        The meta data to store (see `dclab.dfn.config_keys`).
        Each key depicts a meta data section name whose data is given
        as a dictionary, e.g.

            meta = {"imaging": {"exposure time": 20,
                                "flash duration": 2,
                                ...
                                },
                    "setup": {"channel width": 20,
                              "chip region": "channel",
                              ...
                              },
                    ...
                    }

        Only section key names and key values therein registered
        in dclab are allowed and are converted to the pre-defined
        dtype.
    logs: dict of lists
        Each key of `logs` refers to a list of strings that contains
        logging information. Each item in the list can be considered to
        be one line in the log file.
    mode: str
        Defines how the input `data` and `logs` are stored:
        - "append": append new data to existing Datasets; the opened
                    `h5py.File` object is returned (used in real-
                    time data storage)
        - "replace": replace keys given by `data` and `logs`; the
                    opened `h5py.File` object is closed and `None`
                    is returned (used for ancillary feature storage)
        - "reset": do not keep any previous data; the opened
                   `h5py.File` object is closed and `None` is returned
                   (default)
    compression: str
        Compression method for "contour", "image", and "trace" data
        as well as logs; one of [None, "lzf", "gzip", "szip"].

    Notes
    -----
    If `data` is an instance of RTDCBase, then `meta` must be set to
    `data.config`, otherwise no meta data will be saved.
    """
    if mode not in ["append", "replace", "reset"]:
        raise ValueError("`mode` must be one of [append, replace, reset]")
    if compression not in [None, "gzip", "lzf", "szip"]:
        raise ValueError("`compression` must be one of [gzip, lzf, szip]")

    # make sure we are not overriding anything
    meta = copy.deepcopy(meta)

    if (not hasattr(data, "__iter__") or
        not hasattr(data, "__contains__") or
            not hasattr(data, "__getitem__") or
            isinstance(data, (list, np.ndarray))):
        msg = "`data` must be dict-like"
        raise ValueError(msg)

    # Check meta data
    for sec in meta:
        if sec not in dfn.config_keys:
            # only allow writing of meta data that are not editable
            # by the user (not dclab.dfn.CFG_ANALYSIS)
            msg = "Meta data section not defined in dclab: {}".format(sec)
            raise ValueError(msg)
        for ck in meta[sec]:
            if ck not in dfn.config_keys[sec]:
                msg = "Meta key not defined in dclab: {}:{}".format(sec, ck)
                raise ValueError(msg)

    # Check feature keys
    feat_keys = []
    for kk in data:
        if dfn.feature_exists(kk):
            feat_keys.append(kk)
        else:
            raise ValueError("Unknown key '{}'!".format(kk))
        # verify trace names
        if kk == "trace":
            for sk in data["trace"]:
                if sk not in dfn.FLUOR_TRACES:
                    msg = "Unknown trace key: {}".format(sk)
                    raise ValueError(msg)

    # Create file
    # (this should happen after all checks)
    if isinstance(path_or_h5file, h5py.File):
        h5obj = path_or_h5file
    else:
        if mode == "reset":
            h5mode = "w"
        else:
            h5mode = "a"
        h5obj = h5py.File(path_or_h5file, mode=h5mode)

    # update version
    # - if it is not already in the hdf5 file (prevent override)
    # - if it is explicitly given in meta (append to old version string)
    if ("setup:software version" not in h5obj.attrs
            or ("setup" in meta and "software version" in meta["setup"])):
        thisver = "dclab {}".format(version)
        if "setup" in meta and "software version" in meta["setup"]:
            oldver = meta["setup"]["software version"]
            thisver = "{} | {}".format(oldver, thisver)
        if "setup" not in meta:
            meta["setup"] = {}
        meta["setup"]["software version"] = thisver
    # Write meta
    for sec in meta:
        for ck in meta[sec]:
            idk = "{}:{}".format(sec, ck)
            conffunc = dfn.config_funcs[sec][ck]
            h5obj.attrs[idk] = conffunc(meta[sec][ck])

    # Write data
    # create events group
    events = h5obj.require_group("events")
    # remove previous data
    if mode == "replace":
        for rk in feat_keys:
            if rk in events:
                del events[rk]
    # store experimental data
    for fk in feat_keys:
        if dfn.scalar_feature_exists(fk):
            store_scalar(h5group=events,
                         name=fk,
                         data=data[fk],
                         compression=compression)
        elif fk == "contour":
            store_contour(h5group=events,
                          data=data["contour"],
                          compression=compression)
        elif fk == "image":
            store_image(h5group=events,
                        data=data["image"],
                        compression=compression)
        elif fk == "mask":
            store_mask(h5group=events,
                       data=data["mask"],
                       compression=compression)
        elif fk == "trace":
            store_trace(h5group=events,
                        data=data["trace"],
                        compression=compression)

    # Write logs
    if logs:
        log_group = h5obj.require_group("logs")
        # remove previous data
        if mode == "replace":
            for rl in logs:
                if rl in log_group:
                    del log_group[rl]
        dt = h5py.special_dtype(vlen=hdf5_str)
        for lkey in logs:
            ldata = logs[lkey]
            if isinstance(ldata, (str, hdf5_str)):
                # single event
                ldata = [ldata]
            lnum = len(ldata)
            if lkey not in log_group:
                log_dset = log_group.create_dataset(lkey,
                                                    (lnum,),
                                                    dtype=dt,
                                                    maxshape=(None,),
                                                    chunks=True,
                                                    fletcher32=True,
                                                    compression=compression)
                for ii, line in enumerate(ldata):
                    log_dset[ii] = line
            else:
                log_dset = log_group[lkey]
                oldsize = log_dset.shape[0]
                log_dset.resize(oldsize + lnum, axis=0)
                for ii, line in enumerate(ldata):
                    log_dset[oldsize + ii] = line

    if mode == "append":
        return h5obj
    else:
        h5obj.close()
        return None
